from zope import interface

import os
import sys

import utility
import interfaces

IGNORE = object()
DELETE = object()

def root_length(a, b):
    if b.startswith(a):
        return len(a)
    else:
        return 0

def sort_by_path(path, paths):
    return sorted(
        paths, key=lambda syspath: root_length(syspath, path), reverse=True)

def find_zope2_product(path):
    """Check the Zope2 magic Products semi-namespace to see if the
    path is part of a Product."""

    _syspaths = sort_by_path(path, sys.modules["Products"].__path__)
    syspath = _syspaths[0]

    if not path.startswith(syspath):
        return None

    product = path[len(syspath)+1:].split(os.path.sep, 2)[0]

    return "Products." + product


def find_package(syspaths, path):
    """Determine the Python-package where path is located.  If the path is
    not located within the Python sys-path, return ``None``."""

    _syspaths = sort_by_path(path, syspaths)
    syspath = _syspaths[0]

    path = os.path.normpath(path)
    if not path.startswith(syspath):
        if utility.ZOPE_2:
            return find_zope2_product(path)
        return None

    path = path[len(syspath):]

    # convert path to dotted filename
    if path.startswith(os.path.sep):
        path = path[1:]

    return path

class TemplateManagerFactory(object):
    def __init__(self, name):
        self.manager = TemplateManager(name)

    def __call__(self, layer):
        return self.manager

class TemplateManager(object):
    interface.implements(interfaces.ITemplateManager)

    def __init__(self, name):
        self.syspaths = tuple(sys.path)
        self.templates = {}
        self.paths = {}
        self.directories = set()
        self.name = name

    def registerDirectory(self, directory):
        self.directories.add(directory)

        for filename in os.listdir(directory):
            if filename.endswith('.pt'):
                self.paths[filename] = "%s/%s" % (directory, filename)

        for template, filename in self.templates.items():
            if filename is IGNORE:
                del self.templates[template]

    def unregisterDirectory(self, directory):
        self.directories.remove(directory)

        templates = []

        for template, filename in self.templates.items():
            if filename in self.paths:
                templates.append(template)

        for filename in os.listdir(directory):
            if filename in self.paths:
                del self.paths[filename]

        for template in templates:
            self.registerTemplate(template)
            del self.templates[template]
            template.filename = template._filename
            template._v_last_read = False

    def unregisterAllDirectories(self):
        for directory in tuple(self.directories):
            self.unregisterDirectory(directory)

    def registerTemplate(self, template):
        # only register templates that have a filename attribute
        if not hasattr(template, 'filename'):
            return

        # assert that the template is not already registered
        filename = self.templates.get(template)
        if filename is IGNORE:
            return

        # if the template filename matches an override, we're done
        paths = self.paths
        if paths.get(filename) == template.filename:
            return

        # verify that override has not been unregistered
        if filename is not None and filename not in paths:
            template.filename = template._filename
            del self.templates[template]

        # check if an override exists
        path = find_package(self.syspaths, template.filename)
        if path is None:
            # permanently ignore template
            self.templates[template] = IGNORE
            return

        filename = path.replace(os.path.sep, '.')
        if filename in paths:
            path = paths[filename]

            # save original filename
            template._filename = template.filename

            # save template and registry and assign path
            template.filename = path
            self.templates[template] = filename
        else:
            self.templates[template] = IGNORE

        # force cook
        template._v_last_read = False

        return True



from zope import interface
from zope import component

from zope.app.component.hooks import getSite
from zope.publisher.interfaces import IRequest

import zope.security.management
import zope.security.interfaces

import interfaces

try:
    import Acquisition
    ZOPE_2 = True
except:
    ZOPE_2 = False

def getRequest():
    if ZOPE_2:
        # get request by acquisition
        site = getSite()
        if site is not None:
            try:
                return site.request
            except AttributeError:
                return site.REQUEST

    try:
        i = zope.security.management.getInteraction()
    except zope.security.interfaces.NoInteraction:
        return

    for p in i.participations:
        if IRequest.providedBy(p):
            return p

def getLayer():
    request = getRequest()

    if request is not None:
        return interface.providedBy(request)

    return interface.Interface

def getManagers():
    layer = getLayer()
    gsm = component.getGlobalSiteManager()

    for name, factory in gsm.adapters.lookupAll((layer,), interfaces.ITemplateManager):
        yield factory(layer)



from zope.interface import Interface
from zope.configuration import fields
from zope.configuration.fields import GlobalObject

class ITemplateOverridesDirective(Interface):
    """Directive which registers a directory with template overrides."""

    directory = fields.Path(
        title=u"Path to directory",
        required=True)

    layer = GlobalObject(
        title=u"The layer the overrides should be enabled for",
        description=u"By default overrides are used for all layers.",
        required=False
        )




import zope.interface
import zope.component
import zope.testing
import unittest

OPTIONFLAGS = (zope.testing.doctest.ELLIPSIS |
               zope.testing.doctest.NORMALIZE_WHITESPACE)

import zope.component.testing

def test_suite():
    doctests = ['README.txt']
    globs = dict(
        interface=zope.interface,
        component=zope.component)
    
    try:
        import Products.Five
        doctests.append('Five.txt')
    except:
        pass
    
    return unittest.TestSuite((
        zope.testing.doctest.DocFileSuite(doctest,
                                          optionflags=OPTIONFLAGS,
                                          setUp=zope.component.testing.setUp,
                                          tearDown=zope.component.testing.tearDown,
                                          globs=globs,
                                          package="z3c.jbot") for doctest in doctests
        ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from zope import interface
from zope import component

import manager
import interfaces

def handler(directory, layer):
    lookup_all = component.getGlobalSiteManager().adapters.lookupAll

    # check if a template manager already exists
    factories = set(factory for name, factory in lookup_all(
        (layer,), interfaces.ITemplateManager))

    # if factory is available on the interface bases of the layer we
    # discard it and register a new manager specialized to the layer
    if layer is interface.Interface:
        base_factories = set()
    else:
        base_factories = set()
        for base in layer.__bases__:
            for name, factory in lookup_all((base,), interfaces.ITemplateManager):
                base_factories.add(factory)

    try:
        factory = factories.difference(base_factories).pop()
    except KeyError:
        name = directory
        factory = manager.TemplateManagerFactory(name)
        component.provideAdapter(
            factory, (layer,), interfaces.ITemplateManager, name=name)

    factory(layer).registerDirectory(directory)

    return factory(layer)

def templateOverridesDirective(_context, directory, layer=interface.Interface):
    _context.action(
        discriminator = ('override', directory, layer),
        callable = handler,
        args = (directory, layer),
        )



from zope.pagetemplate.pagetemplatefile import PageTemplateFile

import utility
import logging
import threading

logger = logging.getLogger('jbot')

PT_CLASSES = [PageTemplateFile]

try:
    import Products.PageTemplates.PageTemplateFile
    PT_CLASSES.append(Products.PageTemplates.PageTemplateFile.PageTemplateFile)
except:
    pass

_v_cache = threading.local()

class LayerProperty(property):
    """Layer-specific property class.

    Instance attributes are instance *and* layer-specific when defined
    using this property class.

    Lookup order:

      1. By layer
      2. By instance
      3. By class

    This pattern takes into account that attributes may be set before
    the property is defined on the class.
    """

    def __init__(self, cls, name):
        self.name = name
        self.default = getattr(cls, name, None)
        property.__init__(self, self._get, self._set)

    def _get(self, template):
        key = self.name
        __dict__ = template.__dict__

        # note: exceptions should not arise during normal service
        try:
            layer = _v_cache.layer
        except AttributeError:
            layer = None

        try:
            # any of these could raise a key-error
            attrs = __dict__['_v_attrs']
            return attrs[layer, key]
        except KeyError:
            return __dict__.get(key) or self.default

    def _set(self, template, value):
        key = self.name
        __dict__ = template.__dict__

        # note: exceptions should not arise during normal service
        try:
            layer = _v_cache.layer
        except AttributeError:
            layer = None

        try:
            attrs = __dict__['_v_attrs']
        except KeyError:
            attrs = __dict__['_v_attrs'] = {}

        attrs[layer, key] = value
        __dict__.setdefault(key, value)

# registration hook to template manager
def jbot(func):
    def patch(self, *args, **kwargs):
        _v_cache.layer = utility.getLayer()

        for manager in utility.getManagers():
            # register template; this call returns ``True`` if
            # template was invalidated
            if manager.registerTemplate(self):
                break

        return func(self, *args, **kwargs)
    return patch

logger.info("Patching page template classes for use with z3c.jbot...")

# patch ``_cook_check``-method to insert jbot-logic
for pt_class in PT_CLASSES:
    pt_class._cook_check = jbot(pt_class._cook_check)

# munge per-layer attribute descriptors on class
for pt_class in PT_CLASSES:
    for name in ('_v_macros', '_v_program', '_v_cooked', '_v_errors',
                 '_v_last_read', '_v_warning', '_text_',
                 'filename', '_filename', 'content_type', 'is_html'):
        setattr(pt_class, name, LayerProperty(pt_class, name))
